/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is __________________________________________.
 *
 * The Initial Developer of the Original Code is
 * ____________________________________________.
 * Portions created by the Initial Developer are Copyright (C) 2___
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

/* FIXME: must add this file to xpinstall info for each platform */

// Bugzilla Bug 71689 Need helper utilities for JS component authors
//  "They suffer without a generic module implementation, and the
//   cutting-and-pasting breaks my little heart."  (a-yup)
//
// GOAL: no more rope than absolutely necessary to shoot yourself in the foot
//
// NOTE: No include files are allowed for component files, so these
//       utils are added to the environment in which JS compenents are run.
//       All objects in this will become properties on the 'Utils' object
//
// var Utils = {
// 
//    // take a list of one or more components and generate a wrapping module
//    genericModule: function(name, component) 
//
//    // generate a bare bones module that needs only its real interface
//    genericComponent: function(name, cid, contract, isSingleton 
//                               /* interface, ... */) 
//
//    // helper function to add an entry to the category manager
//    addCategoryEntry: function(category, name, value) 
//
//    // helper function to remove an entry from the category manager
//    deleteCategoryEntry: function(category, name) 
//
//    // format and print a debug message 
//    printDebug: function(message)
//
//    // set "Utils.debug" to true to print internal debugging statements
//    var debug = false
// 
//  }

var debug = false;

// wrap a module around a list of one more component objects
function genericModule(name /* , component, ... */) {

  // the first argument is the module name (used for debugging only)
  this.name = name;

  // and the rest are a list of components
  this.componentArray = new Array(); 
  for (var i = 1; i < arguments.length; i++) {
    this.componentArray.push(arguments[i]);
  }

  // use Utils as the default for printing debugging statements
  this.debug = Utils.debug;

  // override printDebug() to make messages slightly more informative
  this.printDebug = function(message) {
    if (this.debug) Utils.printDebug("Module " + this.name + ": " + message);
  }

  if (this.componentArray.length == 0) {
    this.printDebug("(WARNING) no components given to genericModule");
  }

  this.registerSelf = function(compMgr, fileSpec, location, type) {
    this.printDebug("entering registerSelf...");
    var nsIComponentRegistrar = Components.interfaces.nsIComponentRegistrar;
    var registrar = compMgr.QueryInterface(nsIComponentRegistrar);
    
    for (i = 0; i < this.componentArray.length; i++) {
      var component = this.componentArray[i];
      this.printDebug(" registering component '" + component.name + "'");
      registrar.registerFactoryLocation(component.cid,
                                        component.name,
                                        component.contract,
                                        fileSpec,
                                        location,
                                        type);
      // callback to allow component specific registration
      component.register(compMgr, fileSpec, location, type);
    }
    this.printDebug("leaving registerSelf");
  }
  
  this.unregisterSelf = function(compMgr, fileSpec, location) {
    
    this.printDebug("entering unregisterSelf...");
    var nsIComponentRegistrar = Components.interfaces.nsIComponentRegistrar;
    var registrar = compMgr.QueryInterface(nsIComponentRegistrar);
    
    for (i = 0; i < this.componentArray.length; i++) {
      var component = this.componentArray[i];
      this.printDebug(" unregistering component '" + component.name + "'");
      registrar.registerFactoryLocation(component.cid, fileSpec);
      // callback to allow component specific unregistration
      component.unregister(compMgr, fileSpec, location, type);
    }
    this.printDebug("leaving unregisterSelf");
  }

  this.getClassObject = function(compMgr, cid, iid) {
    this.printDebug("getClassObject() called");

    if (!iid.equals(Components.interfaces.nsIFactory)) {
      this.printDebug("(error) iid was '" + iid + "'");
      throw Components.results.NS_ERROR_NOT_IMPLEMENTED;
    }

    for (i = 0; i < this.componentArray.length; i++) {
      var component = this.componentArray[i];

      if (cid.equals(component.cid)) {
        this.printDebug(" MATCHED component '" + component.name + "'!");
        return component;    // component acts as its own nsIFactory
      }
      this.printDebug(" no match for component '" + component.name + "'");
    }

    // no component had a matching contract id
    this.printDebug("(error) unable to match cid='" + cid + "'");
    throw Components.results.NS_ERROR_NO_INTERFACE;
  }
 
  this.canUnload = function(compMgr) {
    this.printDebug("canUnload called");
    return true; // return value is ignored for javascript components
  }
};

// make a bare bones component with all supported interfaces at the top level
function genericComponent(name, cid, contract, 
                          isSingleton /*, interface, ... */) 
{
  this.name = name;
  this.cid = cid;
  this.contract = contract;
  this.isSingleton = isSingleton;

  this.interfaceArray = new Array(); 
  // all components have nsISupports so add it automatically
  this.interfaceArray.push(Components.interfaces.nsISupports);
  // then the remaining arguments are interfaces supported by this component
  for (var i = 4; i < arguments.length; i++) {
    this.interfaceArray.push(arguments[i]);
  }

  // use Utils as the default for printing debugging statements
  this.debug = Utils.debug;

  // override printDebug() to make messages slightly more informative
  this.printDebug = function(message) {
    if (this.debug) Utils.printDebug("Component " +this.name+ ": " + message);
  }

  // check that the caller actually gave an interface to use
  if (this.interfaceArray.length == 1) {
    this.printDebug("(WARNING) no user specified interfaces");
  }

  // warn if it is obvious something else has been passed instead of a CID
  //   (the raw xpconnect error is obscure enough to warrant a warning) 
  if (typeof(cid) != "object") {
    this.printDebug("(WARNING) cid '" + cid + "' is not an object!");
  }

  // register() is a callback from genericModule.registerSelf()
  this.register = function(compMgr, fileSpec, location, type) {
    this.printDebug("register called on component '" + component.name + "'");
    return;  // user overrides 'register()' for component specific registration
  }

  // unregister() is a callback from genericModule.unregisterSelf()
  this.register = function(compMgr, fileSpec, location, type) {
    this.printDebug("unregister called on component '" + component.name + "'");
    return;  // user can override for component specific unregistration
  }

  // QueryInterface is part of nsISupports (required for all XPCOM objects)
  this.QueryInterface = function(iid) {
    this.printDebug("QueryInterface called");
    
    // see if this component object supports the requested interface
    for (i = 0; i < this.interfaceArray.length; i++) {
      if (iid.equals(this.interfaceArray[i])) {
        this.printDebug(" MATCHED interface '" + this.interfaceArray[i] + "'");
        return this;
      }
    }
    
    if (this.debug) { // expensive but rare check to get better debug message
      for (var name in Components.interfaces) {
        if (iid.equals(Components.interfaces[name])) {
          this.printDebug("(error) no match for interface '" + name + "'");
          throw Components.results.NS_ERROR_NO_INTERFACE;
        }
      }
      this.printDebug("(error) no match for interface '" + iid + "'");
    }
    throw Components.results.NS_ERROR_NO_INTERFACE;
  }
  
  // required by nsIFactory so define to prevent warnings
  this.lockFactory = function(lock) {
    this.printDebug("lockFactory() called");
    return true;
  }

  // nsIFactory function pointed to by module.getClassObject()
  this.createInstance = function(outer, iid) {
    this.printDebug("createInstance called");
    
    if (outer != null) {
      /* FIXME: tell me again why don't we handle aggregation? */
      this.printDebug("(error) interface does not allow aggregation");
      throw Components.results.NS_ERROR_NO_AGGREGATION;
    }

    if (this.isSingleton) {
      this.printDebug(" using existing interface object");
      return component.QueryInterface(iid);
    }
    else {
      this.printDebug(" creating new interface instance");
      return new component.QueryInterface(iid);
    }

  }
}

// helper function to add an entry to the category manager
function addCategoryEntry(category, name, value) {
  this.printDebug("addCategory: adding to category '" + category + "'");
  var manager = Components.classes["@mozilla.org/categorymanager;1"];
  var service = manager.getService(Components.interfaces.nsICategoryManager);
  /* NOTE: different categories use different entries for name and value */
  service.addCategoryEntry(category, name, value, true, true);
}
  
// helper function to remove an entry from the category manager
function deleteCategoryEntry(category, name) {
  this.printDebug("deleteCategory: from category '" + category + "'");
  var manager = Components.classes["@mozilla.org/categorymanager;1"];
  var service = manager.getService(Components.interfaces.nsICategoryManager);
  service.deleteCategoryEntry(category, name, true);
}

// print a debug message (turn on and off via Utils.debug)  
function printDebug(message) {
  if (this.debug) dump(message + "\n");
}













